// TDoingAgent.cpp: implementation of the TDoingAgent class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdbnx.h"
#include "Basis.h"
#include "TVector.h"
#include "BasisDoc.h"
#include "BasisView.h"
#include "TPerception.h"
#include "TThisPlayer.h"
#include "TDoingAgent.h"
#include <math.h>

// minimal dash power (for action parametrization) min 0 max 100
#define MIN_DASH	60
#define MAX_SELECTION_TRYS		4
#define DIFF_ANGLE  110

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*-----------------------------------------------------------------------------
Function: TDoingAgent::TDoingAgent(CBasisDoc* pDoc, CBasisView* pBView, TPerceptionView* pPView
								 , TNetworkView* pNView, TRuleView* pRView) 
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: pointer to the document
			  pointer to the view containing network traffic
			  pointer to the view for perception visualization
			  pointer to the view for network visualization
			  pointer to the view for behavior rules
Returnval: (void)
Description: constructor
-----------------------------------------------------------------------------*/
TDoingAgent::TDoingAgent(CBasisDoc* pDoc, CBasisView* pBView, TPerceptionView* pPView
								 , TNetworkView* pNView, TRuleView* pRView) 
		: TViewingAgent(pBView, pPView, pNView, pRView)
{
	m_pDoc = pDoc;
	m_bAgentIsReady = FALSE;
	m_bReceivedSee = FALSE;
	m_bSelectedReflex = FALSE;
	m_iSenseBodyCount = 0;
	
	// for statistics
	m_dAvgDistance = 0.0;
	m_lNoOfRelaxes = 0L;
	m_iKickAngle	= 0;
}

TDoingAgent::~TDoingAgent()
{
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::SetKnowledgeType(int type)
Date    : 19.11.98
-------------------------------------------------------------------------------
Parameter: type of knowledge (MAES or DORER or DORER2)
Returnval: (void)
Description: calls base class function and adds action functions
-----------------------------------------------------------------------------*/
void TDoingAgent::SetKnowledgeType(int type)
{
	// create the network
	TViewingAgent::SetKnowledgeType(type);

	// add the perception functions
	CreateActions();
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::OnReceivedMessage(LPCTSTR message)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the message string received
Returnval: (void)
Description: calls the base class function to parse the message and then calls
				 to update the percept class in knowledge
-----------------------------------------------------------------------------*/
void TDoingAgent::OnReceivedMessage(LPCTSTR message, BOOL bFromNetwork)
{
	if ((message[1] == 'e') && (message[2] == 'r'))
	{
		// an error message is here
//		m_pDoc->OnConnectionError();
		m_bHadConnectionError = TRUE;
	}
	else
	{
		// tell the document, that we have connection
		m_pDoc->SetConnected();
		m_bHadConnectionError = FALSE;
	}

	TViewingAgent::OnReceivedMessage(message, bFromNetwork);

	// update the propositions in knowledge
	m_pBehaviorNetwork->UpdatePerception(bFromNetwork);

	// after sense body, the agent acts
	#ifdef SYNCHRONIZE_ON_SENSE_BODY
		if ((message[1] == 's') && (message[3] == 'n'))
			// it is a sensebody info, sent every 100 ms
			OnTimer();
	#endif
}

// for autorunning games

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::OnReceivedSee()
Date    : 23.06.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: called, when the first see command is perceived. calls the documents
				 AutorunReceivedSee, to tell the soccermonitor, the agent is alive
-----------------------------------------------------------------------------*/
void TDoingAgent::OnReceivedSee()
{
	if (!m_bReceivedSee)
	{
		m_bReceivedSee = TRUE;
		if (m_bAgentIsReady)
			m_pDoc->AutorunAgentReady();
	}
}

void TDoingAgent::OnAgentReady()
{
	if (!m_bAgentIsReady)
	{
		m_bAgentIsReady = TRUE;
		if (m_bReceivedSee)
			m_pDoc->AutorunAgentReady();
	}
}

void TDoingAgent::OnReceivedLastSenseBody()
{
	m_pDoc->AutorunSendResults();
}


void TDoingAgent::ReceivedTerminate()
{
	//SenseBody();
	m_pPerception->SetWaitForLastSenseBody(TRUE);
}

BOOL TDoingAgent::IsAutorun() const
{
	return m_pDoc->IsAutorun();
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::OnTimer()
Date    : 23.03.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: called every 100 ms to calculate the best suitable action
-----------------------------------------------------------------------------*/
void TDoingAgent::OnTimer()
{
	int	iNoOfTrys;

#ifdef SHOW_TIME
	
	LARGE_INTEGER	frequency;
	LARGE_INTEGER	start;
	LARGE_INTEGER	stop;
	double			time;
		
	if (QueryPerformanceFrequency(&frequency))
		QueryPerformanceCounter(&start);
#endif

	// time has proceeded
	m_bSelectedReflex = FALSE;
	// perception is updated in the case of no new real perception
	m_pPerception->UpdatePerception();

	// if the agent is not connected, the network is probably a test network,
	// so e.g. true and false are calculated (without perception)
	if (!IsConnected())
		m_pBehaviorNetwork->UpdatePerception(FALSE);

	// activation is spread at least once
	m_pBehaviorNetwork->DoActivationSpreading();

	// only calculate something, if the agent is connected to the server
	if (!IsConnected())
	{
		m_pNetworkView->RedrawView();
		return;
	}

	// only do action if the agent is running
	if (m_pView->GetRunning())
	{
		// selection of reflexes
		if (DoReflexSelection())
			// a reflex has been executed, so no further action to select
			return;

		// execution of queued actions
		if (ExecuteQueuedAction())
			// a queued action has been executed, so no further actions
			return;

		// the agent has to wait for the next see perception
		if (GetWaitForNextSee())
			return;

		// repeat until an action has been selected
		iNoOfTrys = 0;
		while ((!m_pBehaviorNetwork->DoActionSelection()) && (iNoOfTrys < MAX_SELECTION_TRYS))
		{
			// reduce the threshold
			// and do a further activation spreading
			m_pBehaviorNetwork->ReduceThreshold();
			m_pBehaviorNetwork->DoActivationSpreading();
			iNoOfTrys++;
		}
	}
	
	// reduce the threshold for resources not used
	m_pBehaviorNetwork->ReduceThreshold();

	// redraw network, if it is drawn only after a complete spreading
	if (m_pNetworkView->GetDrawAtEndOfActionSelection())
		m_pNetworkView->RedrawView();


#ifdef SHOW_TIME

	if (QueryPerformanceCounter(&stop))
	{
		time = (double)(stop.LowPart - start.LowPart) / (double) frequency.LowPart;
		TRACE1("acttime: %4.3f\n", time);
	}
#endif
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::DoReflexSelection()
Date    : 16.03.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if an action has been performed
Description: checks the situation for reflexes to be done, such as moving to
				 own half after a goal
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::DoReflexSelection()
{
	int	iPlayMode = m_pPerception->GetPlayMode();

	switch (iPlayMode)
	{
		case BEFORE_KICK_OFF:
			// before kickoff, the player moves to his position
			{
				int	x = GetPosKickOffLeft().x;
				int	y = GetPosKickOffLeft().y;
				int	x2 = x;
				int	y2 = y;
				
				// this test is needed because of a strange behavior of the server's move
				// command: players have to be moved with negative x coordinate, no matter
				// on which side they are playing. The GetIsAt function on the other side uses 
				// the 'real' coordinates of a player and therefor players playing from left to 
				// right have to negate the x coordinate for testing (but not for moving)
				if (GetSide() == RIGHT_TO_LEFT)
				{
					x2 = -x2;
					y2 = -y2;
				}

				if (!m_pPerception->GetIsAt(x2, y2))
				{
					Move(x, y);
					
					// perception gets initialized, because object persistence is no longer
					// valid. Ball and players move far distances in no time
					m_pPerception->Initialize();
				}
				else
				{
					// the agent is ready
					OnAgentReady();
					DoSearchBall();
				}
			}
			return TRUE;	// no more actions have to be taken
		
		case KICK_OFF_L:
		case KICK_IN_L:
		case GOAL_KICK_L:
		case FREE_KICK_L:
		case CORNER_KICK_L:
			// if this is a right team player, and is on position, it can wait
			// until game is play on
			// may be should be not dealt with as reflex
			if (GetSide() != LEFT_TO_RIGHT)
			{
				if (m_pPerception->GetObjectDistance(m_pPerception->GetBall()) < 11.0)
				{
					// the player has to stay 9 m away of ball
					DoSearchBall();
					return TRUE;
				}
			}
			break;

		case KICK_OFF_R:
		case KICK_IN_R:
		case GOAL_KICK_R:
		case FREE_KICK_R:
		case CORNER_KICK_R:
			// if this is a right team player, and is on position, it can wait
			// until game is play on
			// may be should be not dealt with as reflex
			if (GetSide() == LEFT_TO_RIGHT)
			{
				if (m_pPerception->GetObjectDistance(m_pPerception->GetBall()) < 11.0)
				{
					// the player has to stay 9 m away of ball
					DoSearchBall();
					return TRUE;
				}
			}
			break;

		case GOAL_R:
			// after a goal, the player moves to his position
			{
				int	x = GetPosKickOffLeft().x;
				int	y = GetPosKickOffLeft().y;
				
				if (!m_pPerception->GetIsAt(x, y))
					Move(x, y);
				else
					DoSearchBall();
			}
			return TRUE;	// no more actions have to be taken
		
		case GOAL_L:
			// after a goal, the player moves to his position
			{
				int	x = GetPosKickOffRight().x;
				int	y = GetPosKickOffRight().y;
				
				if (!m_pPerception->GetIsAt(x, y))
					Move(x, y);
				else
					DoSearchBall();
			}
			return TRUE;	// no more actions have to be taken

		case OFFSIDE_L:
		case OFFSIDE_R:
			break;
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: Functions passed to the Behavior network
Date    : 08.03.99
-----------------------------------------------------------------------------*/
void TDoingAgent::SetGoalTrackingOn(BOOL bVal)
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		((TDorerNetwork*) m_pBehaviorNetwork)->SetGoalTrackingOn(bVal);
}

int TDoingAgent::GetGoalTrackingOn() const
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		return ((TDorerNetwork*) m_pBehaviorNetwork)->GetGoalTrackingOn();
	else
		return FALSE;
}

void TDoingAgent::SetUseFuzzyValues(BOOL bVal)
{
	m_pBehaviorNetwork->SetUseFuzzyValues(bVal);
}

int TDoingAgent::GetUseFuzzyValues() const
{
	return m_pBehaviorNetwork->GetUseFuzzyValues();
}

void TDoingAgent::SetFuzzyRound(double dVal)
{
	m_pBehaviorNetwork->SetFuzzyRound(dVal);
}

void TDoingAgent::SetAdaptActivationAreaOn(BOOL bVal)
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		((TDorerNetwork*) m_pBehaviorNetwork)->SetAdaptActivationAreaOn(bVal);
}

int TDoingAgent::GetAdaptActivationAreaOn() const
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		return ((TDorerNetwork*) m_pBehaviorNetwork)->GetAdaptActivationAreaOn();
	else
		return FALSE;
}

void TDoingAgent::SetCombo(CComboBox* pCombo, BOOL bWhich)
{
	m_pBehaviorNetwork->SetCombo(pCombo, bWhich);
}

CString TDoingAgent::GetActionPerformedToString()
{
	// returns the last performed action, once the agent is ready
	if (!m_bAgentIsReady)
		return CString("Not Ready");
	else
		return m_pBehaviorNetwork->GetActionPerformedToString();
}

CString TDoingAgent::GetResultsInitial() const
{
	return m_pBehaviorNetwork->GetResultsInitial();
}

CString TDoingAgent::GetResults()
{
	CString	result;
	long		count;

	count = m_pPerception->GetNoOfKick() + m_pPerception->GetNoOfDash()
				+ m_pPerception->GetNoOfTurn();
	result.Format(" %5d", count);
	
	// get the results from knowledge
	result += m_pBehaviorNetwork->GetResults();

	return result;
}

// for blocksworld
BOOL TDoingAgent::IsBlocksworld()
{
	return m_pBehaviorNetwork->IsBlocksworld();
}

BOOL TDoingAgent::PerformCycle(TNetworkParams* params, int iNoOfCycles)
{
	return m_pBehaviorNetwork->PerformCycle(params, iNoOfCycles);
}

void TDoingAgent::StopAgent()
{
	m_pDoc->StopAgent();
}

// higher level actions

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::TurnNeck(int iDir, BOOL bQueue)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: angle to turn neck
			  flag indicating whether the action should be queued or immediately executed (default)
Returnval: (void)
Description: turns the neck of the agent, but only if turn angle is more than 
				 a threshold
-----------------------------------------------------------------------------*/
void TDoingAgent::TurnNeck(int iDir, BOOL bQueue)
{
	// check if head is at maxturn
	TDirection headAngle = m_pPerception->GetHeadAngle();
	TDirection turnDir = iDir + headAngle;
	if (turnDir < -90)
		turnDir = -90 - headAngle;
	else if (turnDir > 90)
		turnDir = 90 - headAngle;
	else
		turnDir = iDir;

	// do not turn neck if angle is too small
	if ((turnDir < -5) || (turnDir > 5))
		CBaseAgent::TurnNeck(turnDir, bQueue);
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexDash(double dist, int iDir)
Date    : 24.02.98
-------------------------------------------------------------------------------
Parameter: distance of the goal
			  relative direction of the goal, where to dash
			  power with which to dash (default is 0, which means, that the power is
			  calculated from the distance
			  flag indicating, whether the player should look at the ball while running
Returnval: (void)
Description: avoids to dash into another agent
				 dashes backward, if it is faster (for short traces)
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexDash(double dist, int iDir, int iPower, BOOL bMindBall)
{
	TPlayer*		pPlayer = m_pPerception->GetPlayerInWay(iDir);
	TDirection	dir = iDir;
	int			deltaDir;
	int			minDir;
	int			opponentDir;
	double		distPlayer;
	double		minDist = 2.0 * PLAYER_SIZE;
	TDirection	deltaBallDir;
	TDirection	headAngle = m_pPerception->GetHeadAngle();
	TDirection	actualTurnDir;

	// if power is not set explicitly, the distance defines the dash power
	if (iPower == 0)
	{
		if (GetOrdinaryGame())
			iPower = 100;
		else 
			iPower = (int) (100.0 * GetLinearFuzzyValue(dist, 0.0, 5.0));
	}

	// avoid to run into other players
	if ((pPlayer != NULL) 
		&& ((distPlayer = m_pPerception->GetObjectDistance(pPlayer)) < dist))
	{
		// there is a player in way, so try to face other direction
		opponentDir = m_pPerception->GetThisPlayer()->GetAngleToBody(pPlayer->GetPosition());
		deltaDir = iDir - opponentDir;
//		deltaDir = iDir - m_pPerception->GetObjectDirection(pPlayer);
		if (distPlayer < 0.7)
			if (deltaDir < 0)
				dir = iDir - 90;
			else
				dir = iDir + 90;
		else if (distPlayer < 1.2)
			if (deltaDir < 0)
				dir = iDir - 30;
			else
				dir = iDir + 30;
		else
			if (deltaDir < 0)
				dir = iDir - 10;
			else
				dir = iDir + 10;
		
		minDir = 15;
	}
	else 
	{
		// no one is in way
		// calculate the preciseness for direction
		if (dist > 8.0)
			minDir = 5;
		else
			minDir = (int) (asin(1.0 / (dist + 1.0)) * RAD2DEG);
	}
	
	// calculate if the agent has to turn and how much
	deltaBallDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall()) - dir;
	actualTurnDir = dir + headAngle; 

	if (bMindBall && ((deltaBallDir < -135) || (deltaBallDir > 135))
		&& (m_pPerception->GetThisPlayer()->GetNum() != GOALIE_ID))
		// players do not dash backward
		bMindBall = FALSE;

	if (bMindBall && ((deltaBallDir < -135) || (deltaBallDir > 135)))
	{
		// only the goalie dashes backward
		deltaBallDir = deltaBallDir - 180 - headAngle;

		// player must dash backwards to keep ball in mind
		actualTurnDir -= 180;
		if ((actualTurnDir < -minDir) || (actualTurnDir > minDir))
			// first turn into direction
			// turn body
			Turn(actualTurnDir);
		else
			// we already head the proper direction
			Dash(-iPower);
	}
	else
	{
		if ((actualTurnDir < -minDir) || (actualTurnDir > minDir))
			// first turn into direction
			// turnbody
			Turn(actualTurnDir);
		else
			Dash(iPower);
	}
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexKick(int iPower, int iDirection)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: power and relative direction in which to kick
Returnval: (void)
Description: avoids to kick onto the agent itself and takes the speed of the 
				 ball into account
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexKick(int iPower, int iDirection, bool bQueue)
{
	TDirection	ballDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	double		ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	TDirection	angle = ballDir - iDirection;
	int			newPower = iPower;
	TDirection	newDir = iDirection;
	TPlayer*		opponent = m_pPerception->GetOpponentAtMe();
	TDirection	opponentDir;

	// !!! use in area
	// no care has to be taken if kicking 'forward'
	if ((angle < -100) || (angle > 100))
	{
		if ((ballDist < 1.0) || (angle < -160) || (angle > 160))
		{
			// ball is behind agent, so bring it to either side depending on opponents near
			if ((opponent != NULL) && (m_pPerception->GetObjectDistance(opponent) < 4.0))
			{
				opponentDir = m_pPerception->GetObjectDirection(opponent) - iDirection;
				if (opponentDir > 0)
					// opponent is right, so dribble ball to left
					DribbleClockwise();
				else
					// opponent is left, so dribble ball to right
					DribbleCounterclockwise();
				return;
			}
			else
			{
				// there is no opponent but may be the agent itself blocks the ball
				if ((ballDist < 0.6) || (angle < -160) || (angle > 160))
				{
					int	iAbsX = (int) m_pPerception->GetThisPlayer()->GetAbsX();
					int	iAbsY = (int) m_pPerception->GetThisPlayer()->GetAbsY();

					if (m_pPerception->GetSide() != LEFT_TO_RIGHT)
					{
						iAbsX = -iAbsX;
						iAbsY = -iAbsY;
					}

					if (iAbsX < 0)
					{
						// in own half always dribble away from own goal
						if (iAbsY > 0)
							DribbleCounterclockwise();
						else
							DribbleClockwise();
					}
					else if (angle > 0)
						DribbleCounterclockwise();
					else
						DribbleClockwise();
					return;
				}
				// else kick direct into the direction wanted
			}
		}
	}

	// take the speed of the ball into account
	AccountBallSpeed(newPower, newDir);
	
	if (newPower < 5)
	{
		// the ball already runs into the desired direction, so run to it
		ComplexDashTo(*m_pPerception->GetFutureBallPosition(2), 80, 0);
	}
	else
		Kick(newPower, newDir, bQueue);	
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::AccountBallSpeed(int& power, TDirection& dir)
Date    : 25.08.98
-------------------------------------------------------------------------------
Parameter: power of kicking
			  relative direction in which to kick
Returnval: (void)
Description: take the relative speed vector of the ball into account as well as 
				 the direction in which to kick
-----------------------------------------------------------------------------*/
void TDoingAgent::AccountBallSpeed(int& power, TDirection& dir)
{
	TBall*			pBall = m_pPerception->GetBall();
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();
	double			dBallDist = pBall->GetDistance() - PLAYER_SIZE - BALL_SIZE;
	double			dDeltaDir = (double) pBall->GetDirection() / 180.0;
	TVector			tBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsDir());
	double			dBallSpeed = tBallSpeed.GetLength();
	TVector			tWantToKick;
	double			dKickLen;
	double			dFactor;

	#ifdef SHOW_DOING_VAR
		TRACE2("account ball speed before: power: %d, dir: %d\n", power, (int)dir);
	#endif
	
	if (dDeltaDir < 0)
		dDeltaDir = -dDeltaDir;

	// take head angle into account
	dir += m_pPerception->GetHeadAngle();

	// calculate the wanted final ball speed
	dFactor = KICK_POWER_RATE * (1.0 - 0.25 * dDeltaDir - 0.25 * dBallDist / KICKABLE_MARGIN);
	tWantToKick.SetVectorByDirection(power * dFactor, (int) dir);
	dKickLen = tWantToKick.GetLength();

	if (power == 100)
	{
		// a full power kick does not slow down the ball if it already goes into correct dir
		// so we assume that the want to kick is ballspeed + 100 power
		tWantToKick *= ((dBallSpeed + 100.0 * KICK_POWER_RATE) / dKickLen);
		tWantToKick -= tBallSpeed;
	}
	else
		tWantToKick -= tBallSpeed; 

	dir = tWantToKick.GetDirection();
	power = (int) (tWantToKick.GetLength() / dFactor);

	if (power > 100)
		power = 100;

	#ifdef SHOW_DOING_VAR
		TRACE3("account ball speed: ball speed: %4.2f, balldir: %d, ballpower: %4.2f\n"
					, pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsDir()).GetLength()
					, (int)tBallSpeed.GetDirection()
					, tBallSpeed.GetLength());

		TRACE2("account ball speed after: power: %d, dir: %d\n", power, (int)dir);
	#endif
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::DribbleCounterclockwise()
Date    : 26.08.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: dribbles ball in counterclockwise direction
-----------------------------------------------------------------------------*/
void TDoingAgent::DribbleCounterclockwise()
{
	double		size = PLAYER_SIZE + BALL_SIZE + 0.15;
	TDirection	ballDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	double		ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	int			power;
	TDirection	dir;

	if (ballDist < size)
	{
		// the ball is very close
		// !!! check if the ball has collided with player and has now zero speed
		dir = ballDir - 90;
		power = 15;

		#ifdef _DEBUG
			TRACE2("%d Ball is close counterclockwise! %4.2f\n", m_pPerception->GetSenseTime(), ballDist);
		#endif
	}
	else
	{
		// kick close aside the player
		// since atan is not quite correct we add 30% to the angle
		dir = ballDir - 180 + (int) (atan(size / ballDist) * RAD2DEG * 1.3);
		power = 30;
	}
	AccountBallSpeed(power, dir);
	Kick(power, dir);	
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::DribbleClockwise()
Date    : 26.08.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: dribbles ball clockwise around the player
-----------------------------------------------------------------------------*/
void TDoingAgent::DribbleClockwise()
{
	double		size = PLAYER_SIZE + BALL_SIZE + 0.15;
	TDirection	ballDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	double		ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	int			power;
	TDirection	dir;

	if (ballDist < size)
	{
		// the ball is very close
		dir = ballDir + 90;
		power = 15;
		#ifdef _DEBUG
			TRACE2("%d Ball is close clockwise! %4.2f\n", m_pPerception->GetSenseTime(), ballDist);
		#endif
	}
	else
	{
		// kick close aside the player
		// since atan is not quite correct we add 30% to the angle
		dir = ballDir + 180 - (int)(asin(size / ballDist) * RAD2DEG);
		power = 30;
	}
	AccountBallSpeed(power, dir);
	Kick(power, dir);	
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexDashAndKick(int iPower, int iDirection)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: power, with which to dash
			  direction in which to dash
Returnval: (void)
Description: handles the dribbling of the ball. It is intentionally slower than
				 dashing without ball.
				 does not handle to dribble around a player by now
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexDashAndKick(int iPower, int iDirection)
{
	// direction is relative
	// player runs with the ball (dribling)
	TBall*	pBall = m_pPerception->GetBall();
	double	dMaxKickDistance = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;
	int		ballDir = m_pPerception->GetObjectDirection(pBall);
	double	ballDist = m_pPerception->GetObjectDistance(pBall);
	int		bodyDir = m_pPerception->GetThisPlayer()->GetAbsBodyDir();
	
	if (ballDist > dMaxKickDistance)
	{
		// run to the side of the ball

		if (ballDir - iDirection < 0)
			// run to the right side of the ball
			iDirection = ballDir + 10;
		else
			// run to the left side of the ball
			iDirection = ballDir - 10;

		ComplexDash(ballDist, iDirection, iPower);
	}
	else
		// kick the ball
		ComplexKick(iPower / 3, iDirection);
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::ComplexDashTo(TVector position, int power, int dist)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: absolute field coordinates, where to run to
			  power, whith which to run (default 100)
			  radius of the region which is to reach
			  flag indicating, whether player should look at ball while running

Returnval: TRUE, if the coordinates already have been reached
Description: runs to the specified absolute place in the field. If the place 
				 has already been reached, no dash is done, but the function returns
				 TRUE, otherwise, the player dashes and returns FALSE
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::ComplexDashTo(TVector& position, int power, int dist, BOOL mindBall)
{
	TDirection	direction;
	double		distance;

	// are we at the specified point ?
	if (m_pPerception->GetIsAt((int) position.GetX(), (int) position.GetY(), dist))
		return TRUE;

	// we are not there so where do we have to go
	distance = m_pPerception->GetThisPlayer()->GetDistanceTo(position);
	direction = m_pPerception->GetThisPlayer()->GetAngleTo(position);
	ComplexDash(distance, direction, power, mindBall);

	return FALSE;
}

BOOL TDoingAgent::ComplexDashTo(TVisibleObject* pObj, int power, int dist, BOOL mindBall)
{
	return ComplexDashTo(pObj->GetPosition(), power, dist, mindBall);
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexKickTo(double x, double y, BOOL bStopThere)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: absolute x and y coordinates of the field
			  flag indicating, whether the ball should stop there (TRUE) or just be
			  kicked in that direction
Returnval: (void)
Description: kicks to the specified coordinates. If bStopThere is true, the kick
				 is softer, so that the ball almost stops at the specified position
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexKickTo(TVector& pos, BOOL bStopThere)
{
	int		power;
	int		direction = m_pPerception->GetThisPlayer()->GetAngleTo(pos);
	double	distance = m_pPerception->GetThisPlayer()->GetDistanceTo(pos);

	if (!bStopThere)
	{
		// the ball does not need to stop
		if (distance < 5.0)
			power = 50;
		else if (distance > 20.0)
			power = 100;
		else
			power = ((int) distance - 5) * 50 / 15 + 50;
	}
	else
	{
		// the ball should almost stop there
		if (distance < 5.0)
			power = 10;
		else if (distance > 40.0)
			power = 100;
		else
			power = ((int) distance - 5) * 100 / 35 + 10;
	}

	ComplexKick(power, direction);
}

void TDoingAgent::ComplexKickTo(TVisibleObject* pObj, BOOL bStopThere)
{
	ComplexKickTo(pObj->GetPosition(), bStopThere);
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::CreateActions()
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: adds the agents behavior function to the network
-----------------------------------------------------------------------------*/
void TDoingAgent::CreateActions()
{
	m_pBehaviorNetwork->AddAction("action", this, (PCallBackAction) DummyAction);
	m_pBehaviorNetwork->AddAction("aNothing", this, (PCallBackAction) DoANothing);
	m_pBehaviorNetwork->AddAction("catchBall", this, (PCallBackAction) DoCatchBall);
	m_pBehaviorNetwork->AddAction("checkOffside", this, (PCallBackAction) DoCheckOffside);
	m_pBehaviorNetwork->AddAction("kickAwayBall", this, (PCallBackAction) DoKickAwayBall);
	m_pBehaviorNetwork->AddAction("kickGoal", this, (PCallBackAction) DoKickGoal);
	m_pBehaviorNetwork->AddAction("lookStraight", this, (PCallBackAction) DoLookStraight);
	m_pBehaviorNetwork->AddAction("lookToBall", this, (PCallBackAction) DoLookToBall);
	m_pBehaviorNetwork->AddAction("mindBall", this, (PCallBackAction) DoMindBall);
	m_pBehaviorNetwork->AddAction("moveToHomePosition", this, (PCallBackAction) DoMoveToHomePosition);
	m_pBehaviorNetwork->AddAction("passBall", this, (PCallBackAction) DoPassBall);
	m_pBehaviorNetwork->AddAction("pushBackward", this, (PCallBackAction) DoPushBackward);
	m_pBehaviorNetwork->AddAction("pushForward", this, (PCallBackAction) DoPushForward);
	m_pBehaviorNetwork->AddAction("relax", this, (PCallBackAction) DoRelax);
	m_pBehaviorNetwork->AddAction("retreat", this, (PCallBackAction) DoRetreat);
	m_pBehaviorNetwork->AddAction("runForwardToBall", this, (PCallBackAction) DoRunForwardToBall);
	m_pBehaviorNetwork->AddAction("runToBall", this, (PCallBackAction) DoRunToBall);
	m_pBehaviorNetwork->AddAction("runToHomePosition", this, (PCallBackAction) DoRunToHomePosition);
	m_pBehaviorNetwork->AddAction("runWithBall", this, (PCallBackAction) DoRunWithBall);
	m_pBehaviorNetwork->AddAction("sayPosition", this, (PCallBackAction) DoSayPosition);
	m_pBehaviorNetwork->AddAction("searchBall", this, (PCallBackAction) DoSearchBall);
	m_pBehaviorNetwork->AddAction("turnAroundBallForKickIn", this, (PCallBackAction) DoTurnAroundBallForKickIn);
}

/*-----------------------------------------------------------------------------
Function: Action Functions (sorted alphabetically !!)
Date    : 16.01.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: following are the action functions, that is the basic cognitively
				 uninterruptable behaviors of the agent
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
does nothing
-----------------------------------------------------------------------------*/
void TDoingAgent::DoANothing(double activation)
{
	// this is for statistics, to be able to tell how much of the 
	// competence actions have been performed by the server.
	//	Turn(0);
}

/*-----------------------------------------------------------------------------
called by the goalie to catch the ball with hands
-----------------------------------------------------------------------------*/
void TDoingAgent::DoCatchBall(double activation)
{
	Catch(m_pPerception->GetThisPlayer()->GetAngleToBody(m_pPerception->GetBall()->GetPosition()));
}

/*-----------------------------------------------------------------------------
checks, if this agent is in offside
-----------------------------------------------------------------------------*/
void TDoingAgent::DoCheckOffside(double activation)
{
	// player does not know about offside, so look left and right
	int			count = m_pPerception->GetCheckOffsideCounter();
	TDirection	turnDir;

	// increment the counter
	m_pPerception->SetCheckOffsideCounter(++count);

	if (count == 1)
	{
		// we can not be sure to have seen complete part of field, so look
		if (GetSide() == LEFT_TO_RIGHT)
			turnDir = - 45 - m_pPerception->GetThisPlayer()->GetAbsDir();
		else
			turnDir = - 135 - m_pPerception->GetThisPlayer()->GetAbsDir();

		// turnhead
		Turn(turnDir);
		SetWaitForNextSee();
	}
	else if (count == 2)
	{
		// we can not be sure to have seen complete part of field, so look 90
		if (GetSide() == LEFT_TO_RIGHT)
			turnDir = 45 - m_pPerception->GetThisPlayer()->GetAbsDir();
		else
			turnDir = 135 - m_pPerception->GetThisPlayer()->GetAbsDir();

		// turnhead
		Turn(turnDir);
		SetWaitForNextSee();
	}
	else
	{
		// player now should know to be in offside or not
		m_pPerception->SetCheckOffsideCounter(0);
		if (m_pPerception->KnowOffside())
			DoRetreat(activation);
		else
			DoRunToHomePosition(activation);
	}
}

/*-----------------------------------------------------------------------------
kicks away the ball. Depending on the position of the agent, the ball is
kicked into goal direction, along the lines or away from the own goal to the side
-----------------------------------------------------------------------------*/
void TDoingAgent::DoKickAwayBall(double activation)
{
	// test the direction to the left and right
	TThisPlayer*	thisPlayer = m_pPerception->GetThisPlayer();
	int				side = m_pPerception->GetSide();
	int				x = (int) thisPlayer->GetAbsX();
	int				y = (int) thisPlayer->GetAbsY();
	TDirection		direction;
	TDirection		dir;
	int				iNoOfOpponents;
	int				iMinNoOfOpponents;

	if (side == RIGHT_TO_LEFT)
	{
		// switch the coordinates of the player
		x = -x;
		y = -y;
	}

	if (x > 30)
	{
		DoKickGoal(activation);
		return;
	}
	else if (x > -10)
		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
	else if (y < -14)
		// play along the line
		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftPenaltyCorner());
	else if (y > 14)
		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightPenaltyCorner());
	else if (y < 0)
			direction = thisPlayer->GetRelativeFromAbsolutePlayDirection(-70);
	else
			direction = thisPlayer->GetRelativeFromAbsolutePlayDirection(70);
	
	// see, if there is an opponent in this direction and 10 degrees to the left and right
	dir = direction;

	// !! loop !
	iMinNoOfOpponents = m_pPerception->GetOpponentsWithin(direction - 5, direction + 5);
	iNoOfOpponents = m_pPerception->GetOpponentsWithin(direction + 5, direction + 15);
	if (iNoOfOpponents < iMinNoOfOpponents)
	{
		iMinNoOfOpponents = iNoOfOpponents;
		dir = direction + 10;
	}
	iNoOfOpponents = m_pPerception->GetOpponentsWithin(direction + 15, direction + 25);
	if (iNoOfOpponents < iMinNoOfOpponents)
	{
		iMinNoOfOpponents = iNoOfOpponents;
		dir = direction + 20;
	}
	iNoOfOpponents = m_pPerception->GetOpponentsWithin(direction - 15, direction - 5);
	if (iNoOfOpponents < iMinNoOfOpponents)
	{
		iMinNoOfOpponents = iNoOfOpponents;
		dir = direction - 10;
	}
	iNoOfOpponents = m_pPerception->GetOpponentsWithin(direction - 25, direction - 15);
	if (iNoOfOpponents < iMinNoOfOpponents)
		dir = direction - 20;

	ComplexKick(100, dir);
}

/*-----------------------------------------------------------------------------
kicks the ball into that half of the goal, where less players are
-----------------------------------------------------------------------------*/
void TDoingAgent::DoKickGoal(double activation)
{
	TPlayer*		pOpponent = m_pPerception->GetOpponentAtBall();
	int			dirLeft = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftGoalPost());
	int			dirMitte = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
	int			dirRight = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightGoalPost());
	TDirection	deltaLeft = dirMitte - dirLeft;
	TDirection	deltaRight = dirRight - dirMitte;
	TDirection	dirKick = dirMitte;

	int			playersLeft = m_pPerception->GetPlayersWithin(dirLeft, dirMitte);
	int			playersRight = m_pPerception->GetPlayersWithin(dirMitte, dirRight);

	if (playersLeft < playersRight)
		// if there is anyone, kick into left half of goal
		dirKick = dirLeft + (int) (deltaLeft) / 3;
	else if (playersLeft > playersRight)
		// kick into right half
		dirKick = dirRight - (int) (deltaRight) / 3;
	else if (deltaLeft > deltaRight)
		// kick to the wider half of the goal
		dirKick = dirLeft + (int) (deltaLeft) / 3;
	else
		// kick to the wider half of the goal
		dirKick = dirRight - (int) (deltaRight) / 3;

	if ((pOpponent == NULL) || (m_pPerception->GetObjectDistance(pOpponent) > 8.0))
	{
		// there is no opponent near, so we can put the ball properly to the side
		double	dMaxKick = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;
		TBall*	pBall = m_pPerception->GetBall();
		TVector	tFutureBall = *m_pPerception->GetFutureBallPosition(1);
		int		iHeadAngle = m_pPerception->GetHeadAngle();

		// if the ball will stay close, we can first look if it is good to turn into kick
		// kick direction
		if (m_pPerception->GetThisPlayer()->GetDistanceTo(tFutureBall) < dMaxKick)
		{
			// we have time to do turn or dash
			TDirection turnDir = dirMitte + iHeadAngle;
			if ((turnDir < -30) || (turnDir > 30))
			{
				// turn into kick direction
				Turn(turnDir);
				return;
			}
			else
			{
				// no turn needed, perhaps a little dash
				int	ballDir = m_pPerception->GetThisPlayer()->GetAngleTo(tFutureBall);

				if ((ballDir > -45) && (ballDir < 45) 
						&& (m_pPerception->GetObjectDistance(pBall) > PLAYER_SIZE + BALL_SIZE + 0.2))
				{
					Dash(70);
					return;
				}
			}
		}
	}

	ComplexKick(100, dirMitte);
}

/*-----------------------------------------------------------------------------
turns head into running direction
-----------------------------------------------------------------------------*/
void TDoingAgent::DoLookStraight(double activation)
{
	int	headAngle = m_pPerception->GetHeadAngle();

	if (headAngle != 0)
		TurnNeck(-headAngle);
}

/*-----------------------------------------------------------------------------
this could implement a behavior, where the agent has got enough time to
look around every while
-----------------------------------------------------------------------------*/
void TDoingAgent::DoLookToBall(double activation)
{
	// !!! used as long there is no better implementation
	DoSearchBall();
}

/*-----------------------------------------------------------------------------
keeps the eyes pointed to the ball if possible else into rundirection
-----------------------------------------------------------------------------*/
void TDoingAgent::DoMindBall(double activation)
{
	TDirection	ballDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	int			headAngle = m_pPerception->GetHeadAngle();
	TDirection	deltaDir = ballDir + headAngle;

	if ((deltaDir < -135) || (deltaDir > 135))
		// the head can not be turned
		DoLookStraight(activation);
	else
		TurnNeck(ballDir);
}

/*-----------------------------------------------------------------------------
the goalie can move to its home position after a catch command
-----------------------------------------------------------------------------*/
void TDoingAgent::DoMoveToHomePosition(double activation)
{
	int			ballY = (int) m_pPerception->GetBall()->GetAbsY();
	TDirection	dir;

	// move to 5m corner
	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if (ballY > 0)
		{
			// left top 5m corner
			Move(-48, -9);
			dir = m_pPerception->GetThisPlayer()->GetRelativeFromAbsolutePlayDirection(-70);
		}	
		else
		{
			// left bottom 5m corner
			Move(-48, 9);
			dir = m_pPerception->GetThisPlayer()->GetRelativeFromAbsolutePlayDirection(70);
		}
	}
	else
	{
		if (ballY > 0)
		{
			// right top 5m corner
			Move(-48, 9);
			dir = m_pPerception->GetThisPlayer()->GetRelativeFromAbsolutePlayDirection(70);
		}
		else
		{
			// right bottom 5m corner
			Move(-48, -9);
			dir = m_pPerception->GetThisPlayer()->GetRelativeFromAbsolutePlayDirection(-70);
		}
	}

	// then turn into the kickaway direction (remember that the ball is  now on the other side)
	m_pPerception->SetSpeed(0.0);
	Turn(dir, TRUE);

	// update ball
	m_pPerception->SetGoalieCaught(FALSE);
	m_pPerception->GetBall()->SetPosition(m_pPerception->GetThisPlayer()->GetPosition()
														+ TVector(0.1, 0.0));
	m_pPerception->GetBall()->SetAbsSpeed(TVector(0.0, 0.0));

	// wait a few cycles
	Turn(0, TRUE);
	Turn(0, TRUE);
	Turn(0, TRUE);
	Turn(0, TRUE);

	// and prevent the agent from catching again
	m_pPerception->SetLastCatchTime(m_pPerception->GetSenseTime() + 8);
}

/*-----------------------------------------------------------------------------
kicks the ball direct to the free teammate
-----------------------------------------------------------------------------*/
void TDoingAgent::DoPassBall(double activation)
{
	TPlayer*	pTeammate = m_pPerception->GetFreeTeammate();
	
	if (pTeammate != NULL)
	{
		ComplexKickTo(pTeammate);
	}
	else
		// there is no free teammate (should not happen)
		DoRunWithBall(activation);
}

/*-----------------------------------------------------------------------------
defenders call this to move the offside trap backward
-----------------------------------------------------------------------------*/
void TDoingAgent::DoPushBackward(double activation)
{
	// going back by defenders
	int	x = GetPosHome().x;
	int	y = GetPosHome().y;
	int	x2 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	TPlayer*	teammate;
	int	xDefender;
	int	xBall;
	int	number = m_pPerception->GetThisPlayer()->GetNum();
	int	border;
	int	power;

	if (number <= MAX_DEFENDER)
		// defender
		border = 40;
	else
	{
		// midfielder keeps 10m before defender or ball
		xBall = (int) m_pPerception->GetBall()->GetAbsX();
		teammate = m_pPerception->Get2BackmostTeammate();
		if (teammate != NULL)
			xDefender = (int) teammate->GetAbsX();
		else
			xDefender = xBall;

		if (GetSide() == LEFT_TO_RIGHT)
		{
			xDefender = -xDefender;
			xBall = -xBall;
		}
		if (xBall > xDefender)
			// if the ball is closer to the own goal than our defenders, push backwards
			xDefender = xBall;

		border = xDefender - 10;
		if (border > 20)
			// midfielder maximally push back to x = 20
			border = 20;
	}

	if (GetSide() == LEFT_TO_RIGHT)
	{
		// push back up to border depends on defender or midfielder
		if (x2 > -border)
			// move home pos backward 
			SetPosHome(CPoint((x = x2 - 6), y));
	}
	else
	{
		// push back to the left up to 40
		if (x2 < border)
			// move home pos backward
			SetPosHome(CPoint((x = x2 + 6), y));
	}

	if (m_pPerception->GetStamina() > 2500)
		power = 60;
	else if (m_pPerception->GetStamina() > 1500)
		power = 40;
	else
		power = 10;

	// run to (new) home position
	if (ComplexDashTo(TVector(x, y), power, 3, TRUE))
		// agent is already there
		DoLookToBall(activation);
}

/*-----------------------------------------------------------------------------
called to build up an offside trap by defenders
-----------------------------------------------------------------------------*/
void TDoingAgent::DoPushForward(double activation)
{
	int	x = GetPosHome().x;
	int	y = GetPosHome().y;
	int	x2 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	number = m_pPerception->GetThisPlayer()->GetNum();
	int	border;
	int	power;

	if (number <= MAX_DEFENDER)
		// defender
		border = 5;
	else
		// midfielder
		border = -15;


	if (GetSide() == LEFT_TO_RIGHT)
	{
		// push to the right up to x = -5 (almost midline) but at most up to the ball - 10
		if (x2 < -border)
			// move home pos forward
			SetPosHome(CPoint((x = x2 + 6), y));
	}
	else
	{
		if (x2 > border)
			// move home pos forward
			SetPosHome(CPoint((x = x2 - 6), y));
	}

	if (m_pPerception->GetStamina() > 2000)
		power = 30;
	else
		power = 10;

	// run to (new) home position
	if (ComplexDashTo(TVector(x, y), power, 3, TRUE))
		// agent is already there
		DoSearchBall();
}

/*-----------------------------------------------------------------------------
does nothing to get more stamina
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRelax(double activation)
{
//	int	i;

//	Turn(0);

	// some statistics about relaxing
	// mean distance to the nearest goal
/*	double dist1 = m_pPerception->GetObjectDistance(m_pPerception->GetOwnGoal());
	double dist2 = m_pPerception->GetObjectDistance(m_pPerception->GetOtherGoal());

	if (dist2 < dist1)
		dist1 = dist2;
	m_dAvgDistance = (m_dAvgDistance * m_lNoOfRelaxes + dist1) / (double)(m_lNoOfRelaxes + 1L);
	m_lNoOfRelaxes += 1L;
*/
	DoLookToBall(activation);
}

/*-----------------------------------------------------------------------------
an offensive player goes backward to leave offside
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRetreat(double activation)
{
	// player is in offside, so go back
	int	x = GetPosHome().x;
	int	y = GetPosHome().y;
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	y1 = (int) m_pPerception->GetThisPlayer()->GetAbsY();
	int	power;

	// !!! the agent has to check, if he is out of offside

	if (GetSide() == LEFT_TO_RIGHT)
	{
		if (x > -3)
			// agent is on other side
			SetPosHome(CPoint((x -= 5), y));
	}
	else if (x < 3)
			SetPosHome(CPoint((x += 5), y));

	if (m_pPerception->GetStamina() > 2500)
		power = 60;
	else
		power = 30;

	// run to (new) home position
	if (ComplexDashTo(TVector(x, y), power, 3, TRUE))
		// agent is already there
		DoLookToBall(activation);
}

/*-----------------------------------------------------------------------------
behavior for offenders to keep on the height of the ball
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunForwardToBall(double activation)
{
	int	x = GetPosHome().x;
	int	y = GetPosHome().y;
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	x2 = (int) m_pPerception->GetBall()->GetAbsX();
	int	power;

	if (GetSide() == LEFT_TO_RIGHT)
	{
		if ((x2 > x1 + 5) && (x1 < 45))
			// agent is behind ball
			SetPosHome(CPoint((x = x1 + 5), y));
	}
	else if ((x2 < x1 - 5) && (x1 > -45))
			SetPosHome(CPoint((x = x1 - 5), y));

	// run to (new) home position
	if (m_pPerception->GetStamina() > 2500)
		power = 60;
	else
		power = 30;
	
	if (ComplexDashTo(TVector(x, y), power, 3, TRUE))
		// agent is already there
		DoLookToBall(activation);
}

/*-----------------------------------------------------------------------------
player runs to the expected position of the ball
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunToBall(double activation)
{
	int				i;
	int				delay;
	int				power;
	double			dActionRadius;
	double			dBallDistance;
	TVector			tFutureBall;
	TVector			tPrevious;
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();;
	BOOL				bReady = FALSE;
		
	// calculate the power
	if (GetOrdinaryGame())
		// players in dis games dash parameterized
		power = MIN_DASH + (int) (activation * (100 - MIN_DASH));
	else
		// we dash what we can
		power = 100;
	
	// dash less if freekick
	if ((m_pPerception->GetPlayMode() != PLAY_ON) && !GetOrdinaryGame())
	{
		if (m_pPerception->GetStamina() < LOW_STAMINA)
		{
			// we have too low stamina, so do not dash at all
			DoLookToBall(activation);
			return;
		}
		else
			// dash slow, it is a freekick
			power = 15;
	}

	// calculate direction where to run to
	// delay are the number of cycles that may get lost by turns and other things
	// the bigger delay, the more cautious are the expectations of the agent
	delay = 3;
	i = delay;
	tPrevious = m_pPerception->GetBall()->GetPosition();
	dActionRadius = pAgent->GetFutureDistanceActive(0, power, pAgent->GetAbsDir()
																	, m_pPerception->GetEffort());
	while (!bReady && (i < MAX_FUTURE_POSITIONS))
	{
		// we calculate the intersection of the future agent position and ball position
		tFutureBall = *(m_pPerception->GetFutureBallPosition(i));
		dBallDistance = pAgent->GetDistanceTo(tFutureBall);

		// test if the ball will be outside
		if (m_pPerception->IsOutside(tFutureBall))
		{
			// we run to the last position that is within the field
			// !!! may be if the two distances are big, do not run anyhow
			tFutureBall = tPrevious;
			bReady = TRUE;
		}
			
		else
		{
			// get distance the agent can dash
			dActionRadius = pAgent->GetNextFutureDistance();
			
			if (dActionRadius > dBallDistance)
				// we have the position
				bReady = TRUE;
			else
			{
				// ball will still be too far this setp
				tPrevious = tFutureBall;
				i++;
			}
		}
	}
	
	// execute the dash command
	if (ComplexDashTo(tFutureBall, power, 1, TRUE))
	{
		// agent is already there
/*		if (GetOrdinaryGame())
			// it is a game of the dis
			Turn(0);
		else
*/
		ComplexDashAndKick(100, m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal()));
	}
}

/*-----------------------------------------------------------------------------
runs to the current home position of the player. This position may be specified
by other behaviors
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunToHomePosition(double activation)
{
	int			power;
	int			precision = 3;
	TDirection	dir;
	TVector		posHome;

	if (m_pPerception->GetThisPlayer()->GetNum() == GOALIE_ID)
	{
		// goalie dashes back faster
		power = 60;
		precision = 2;
		
		// and checks, whether to move left or right
		dir = m_pPerception->GetObjectDirection(m_pPerception->GetBall())
				- m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
		
		// check, if the goalie should be in a different position
		switch (m_iIndexPosHome)
		{
		case 0:
				if (dir < 45)
					m_iIndexPosHome = 1;
			break;
		case 1:
				if (dir > 55)
					m_iIndexPosHome = 0;
				else if (dir < 25)
					m_iIndexPosHome = 2;
			break;
		case 3:
				if (dir < -55)
					m_iIndexPosHome = 4;
				else if (dir > -25)
					m_iIndexPosHome = 2;
			break;
		case 4:
				if (dir > -45)
					m_iIndexPosHome = 3;
			break;
		default:
				if (dir > 35)
					m_iIndexPosHome = 1;
				else if (dir < -35)
					m_iIndexPosHome = 3;
			break;
		}

		switch(m_iIndexPosHome)
		{
		case 0:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-50, 7));
			else
				SetPosHome(CPoint(50, -7));
			break;
		case 1:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-48, 4));
			else
				SetPosHome(CPoint(48, -4));
			break;
		case 3:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-48, -4));
			else
				SetPosHome(CPoint(48, 4));
			break;
		case 4:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-50, -7));
			else
				SetPosHome(CPoint(50, 7));
			break;
		default:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-48, 0));
			else
				SetPosHome(CPoint(48, 0));
			break;
		}
	}
	else
		power = 20;

	posHome = TVector(GetPosHome().x, GetPosHome().y);

	if (ComplexDashTo(posHome, power, precision, TRUE))
	{
		// agent is already there
		DoSearchBall(activation);
	}
}

/*-----------------------------------------------------------------------------
the player does a dribbling
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunWithBall(double activation)
{
	int		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
	int		playerDirection;
	TPlayer*	pPlayer = m_pPerception->GetPlayerInWay(direction);

	if (pPlayer == NULL)
		ComplexDashAndKick(100, direction);
	else
	{
		// avoid to dribble into an other agent
		// !!! should be part of the behavior network
		playerDirection = m_pPerception->GetObjectDirection(pPlayer);
		if ((playerDirection > -40) && (playerDirection < 40))
		{
			// when facing an opponent, we kick away the ball or pass it to other agent
			pPlayer = m_pPerception->GetFreeTeammate();
			if ((pPlayer != NULL) && (pPlayer->GetFreePassDegree() > 0.8))
				DoPassBall(activation);
			else
				DoKickAwayBall(activation);
		}
		else
			ComplexDashAndKick(100, direction);
	}

//	DribbleClockwise();
}

/*-----------------------------------------------------------------------------
say the own position and that of the valid ball and all valid players
-----------------------------------------------------------------------------*/
void TDoingAgent::DoSayPosition(double activation)
{
	TThisPlayer *pThisPlayer = m_pPerception->GetThisPlayer();
	CString	sMsg;
	CString	sStr1;
	CString	sStr2;
	char		cSep = ' ';
	int		i;

	sMsg = "*";
	sMsg += cSep;

	//recipient of the message
	i = 0;	//all players
	sStr1.Format("%d",i);
	sMsg += sStr1;
	sMsg += cSep;

	//dispatcher of the message
	i = pThisPlayer->GetNum();
	sStr1.Format("%d",i);
	sMsg += sStr1;
	sMsg += cSep;

	//dispatchers stamina
	i = pThisPlayer->GetStamina();
	sStr1.Format("%d",i);
	sMsg += sStr1;
	sMsg += cSep;

	//msg type p for position
	sMsg += "P ";

	//string of valid ball and players
	sStr1 = m_pPerception->GetValidSeen();
	sMsg += sStr1;

	Say(sMsg);

}


/*-----------------------------------------------------------------------------
searches the ball by looking around and changing view angle to 180
-----------------------------------------------------------------------------*/
void TDoingAgent::DoSearchBall(double activation)
{
	TBall*		pBall = m_pPerception->GetBall();
	TDirection	headAngle = m_pPerception->GetHeadAngle();
	
	if (pBall->IsInMind())
	{
		// it is allready seen or at least in mind so turn to the direction
		int	deltaDir = m_pPerception->GetObjectDirection(pBall) + headAngle;
		if ((deltaDir > 8) || (deltaDir < -8))
			// turnhead
			Turn(deltaDir);
	}
	else 
	{
		// we have no idea, where the ball should be
		// turnhead
		Turn(90);
		SetWaitForNextSee();
	}
}

/*-----------------------------------------------------------------------------
the player runs behind the ball if playmode is KICK_IN
-----------------------------------------------------------------------------*/
void TDoingAgent::DoTurnAroundBallForKickIn(double activation)
{
	// test the direction to the left and right
	double MIN_DISTANCE_TO_BALL = 3.0;
	int KICK_IN_DASH_POWER = 30;
	int MIN_DISTANCE_TO_POSITION = 2;
	TBall*			pBall = m_pPerception->GetBall();
	TVector			posBall = pBall->GetPosition();

	if (posBall.GetY() > 0)
	{
		posBall += TVector(0.0, MIN_DISTANCE_TO_BALL);
		if (ComplexDashTo(posBall,	KICK_IN_DASH_POWER, MIN_DISTANCE_TO_POSITION))
			DoRunToBall(activation);
	}
	else
	{
		posBall -= TVector (0.0, MIN_DISTANCE_TO_BALL);
		if (ComplexDashTo(posBall,	KICK_IN_DASH_POWER, MIN_DISTANCE_TO_POSITION))
			DoRunToBall(activation);
	}
}


