// TViewingAgent.cpp: implementation of the TViewingAgent class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdbnx.h"
#include "Basis.h"
#include "TDirection.h"
#include "BasisView.h"
#include "TScreenParams.h"
#include "TPerception.h"
#include "TThisPlayer.h"
#include "TViewingAgent.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*-----------------------------------------------------------------------------
Function: TViewingAgent::TViewingAgent(CBasisView* pBView, TPerceptionView* pPView
									  , TNetworkView* pNView, TRuleView* pRView)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: pointer to the view containing network traffic
			  pointer to the view for perception visualization
			  pointer to the view for network visualization
			  pointer to the view for behavior rules
Returnval: (void)
Description: constructor
-----------------------------------------------------------------------------*/
TViewingAgent::TViewingAgent(CBasisView* pBView, TPerceptionView* pPView
									  , TNetworkView* pNView, TRuleView* pRView)
{
	m_pView = pBView;
	m_pNetworkView = pNView;
	m_pRuleView = pRView;
	m_pPerception = new TPerception(this, pPView);
	m_iPosKickOffLeft = CPoint(0,0);
	m_iPosKickOffRight = CPoint(0,0);
	m_bOrdinaryGame = FALSE;
	m_bWaitForNextSee = FALSE;
	m_pBehaviorNetwork = NULL;
	m_iIndexPosHome = 2;
}

TViewingAgent::~TViewingAgent()
{
	delete m_pBehaviorNetwork;
	delete m_pPerception;
}


/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SetKnowledgeType(int type)
Date    : 19.11.98
-------------------------------------------------------------------------------
Parameter: type of knowledge (MAES or DORER or DORER2)
Returnval: (void)
Description: deletes the old network and creates a new network
-----------------------------------------------------------------------------*/
void TViewingAgent::SetKnowledgeType(int type)
{
	if (m_pBehaviorNetwork != NULL)
		delete m_pBehaviorNetwork;

	m_iKnowledgeType = type;
	switch(type)
	{
	case MAES_KNOWLEDGE:
		m_pBehaviorNetwork = new TMaesNetwork(m_pNetworkView, m_pRuleView);
		break;

	case DORER_KNOWLEDGE:
		m_pBehaviorNetwork = new TDorerNetwork(m_pNetworkView, m_pRuleView);
		break;

	case DORER2_KNOWLEDGE:
		m_pBehaviorNetwork = new TDorer2Network(m_pNetworkView, m_pRuleView);
		break;

	default:
	#ifdef _DEBUG
		TRACE("Knowledge type not existing\r\n");
	#endif
		ASSERT(FALSE);		
	}

	// add the perception functions
	CreatePerceptions();

	// add resource functions
	CreateResources();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SetViewParams(int angle, int quality)
Date    : 20.01.98
-------------------------------------------------------------------------------
Parameter: view angle (one of 180, 90 or 45)
			  view quality (0 for low, 1 for high)
Returnval: void
Description: sets the internal representation of the viewing angle and quality
-----------------------------------------------------------------------------*/
void TViewingAgent::OnChangedView(int angle, int quality)
{ 
	m_pPerception->SetViewParams(angle, quality); 
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SwitchViewQuality()
Date    : 19.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: switches the view quality from high to low
-----------------------------------------------------------------------------*/
void TViewingAgent::SwitchViewQuality()
{
	int iQuality = m_pPerception->SwitchViewQuality();
	int iAngle = m_pPerception->GetViewAngle();
	ChangeView(iAngle, iQuality);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::SwitchViewAngle()
Date    : 19.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: switches from a view of 180 to 90 to 45 degrees
-----------------------------------------------------------------------------*/
void TViewingAgent::SwitchViewAngle()
{
	int iAngle = m_pPerception->SwitchViewAngle();
	int iQuality = m_pPerception->GetViewQuality();
	ChangeView(iAngle, iQuality);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SetNumber(int num)
Date    : 17.03.98
-------------------------------------------------------------------------------
Parameter: number of the player
			  side the player is playing (LEFT_TO_RIGHT or RIGHT_TO_LEFT)
Returnval: (void)
Description: sets the number of the player and its position before playon
				 it is assumed, that team has been set before
-----------------------------------------------------------------------------*/
void TViewingAgent::SetNumber(int number, int side)
{ 
	int	posLX[PLAYERS_PER_TEAM] = POS_KICKOFF_LEFT_X;
	int	posLY[PLAYERS_PER_TEAM] = POS_KICKOFF_LEFT_Y;
	int	posRX[PLAYERS_PER_TEAM] = POS_KICKOFF_RIGHT_X;
	int	posRY[PLAYERS_PER_TEAM] = POS_KICKOFF_RIGHT_Y;
	int	homeX[PLAYERS_PER_TEAM] = POS_HOME_X;
	int	homeY[PLAYERS_PER_TEAM] = POS_HOME_Y;

	if (side == LEFT_TO_RIGHT)
	{
		// set the home position left to right player
		m_iPosHome.x = homeX[number - 1];
		m_iPosHome.y = homeY[number - 1];
	}
	else
	{
		// set the home position for a right to left player
		m_iPosHome.x = -homeX[number - 1];
		m_iPosHome.y = -homeY[number - 1];
	}

	// set the positions of the player
	if (side == LEFT_TO_RIGHT)
	{
		// it is a player of the left team
		m_iPosKickOffLeft.x = posLX[number - 1];
		m_iPosKickOffLeft.y = posLY[number - 1];
		m_iPosKickOffRight.x = posRX[number - 1];
		m_iPosKickOffRight.y = posRY[number - 1];
	}
	else
	{
		// it is a player of the right team, so mirror x coordinates
		m_iPosKickOffLeft.x = posRX[number - 1];
		m_iPosKickOffLeft.y = posRY[number - 1];
		m_iPosKickOffRight.x = posLX[number - 1];
		m_iPosKickOffRight.y = posLY[number - 1];
	}

	CBaseAgent::SetNumber(number, side);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::Turn(int iMoment, BOOL bQueue)
Date    : 07.09.98
-------------------------------------------------------------------------------
Parameter: relative angle to turn ranging from -180 to 180 where -90 is 90 degrees to the left
			  flag indicating, whether the action should be performed instantaniously
			  (FALSE, default) or queued and executed next timestep
Returnval: (void)
Description: takes into accoun the speed of the agent. the agent actually turns
				 less then iMoment, if it is fast. so this function increases turn moment
-----------------------------------------------------------------------------*/
void TViewingAgent::Turn(int iMoment, BOOL bQueue)
{
	double		speed = m_pPerception->GetSpeed();

	// increase turn
	double	dTurn = iMoment;
	dTurn *= 1.0 + speed * INERTIA_MOMENT;
	iMoment = (int)dTurn;
	if (iMoment > 180)
		iMoment = 180;
	else if (iMoment < -180)
		iMoment = -180;

	CBaseAgent::Turn(iMoment, bQueue);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::TurnBody(int iMoment, BOOL bQueue)
Date    : 18.06.99
-------------------------------------------------------------------------------
Parameter: relative angle to turn the body of the agent 
			  flag indicating, whether the action should be performed instantaniously
			  (FALSE, default) or queued and executed next timestep
Returnval: (void)
Description: takes into accoun the speed of the agent. the agent actually turns
				 less then iMoment, if it is fast. so this function increases turn moment
-----------------------------------------------------------------------------*/
void TViewingAgent::TurnBody(int iMoment, BOOL bQueue)
{
	TDirection	iTurn(iMoment + m_pPerception->GetHeadAngle());
	Turn(iTurn, bQueue);
}

/*-----------------------------------------------------------------------------
Function: BOOL TViewingAgent::OnConnected()
Date    : 11.05.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: called after the client connected to the server. Starts the timeout
				 timer for automatical shutdown, if the server does not send messages
-----------------------------------------------------------------------------*/
void TViewingAgent::OnConnected()
{
	CBaseAgent::OnConnected();
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::OnDisconnect()
Date    : 11.05.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: sends the bye message to the server
-----------------------------------------------------------------------------*/
void TViewingAgent::OnDisconnect()
{
	if ((m_pView != NULL) && ::IsWindow(m_pView->m_hWnd))
		m_pView->StopTimeoutTimer();
	CBaseAgent::OnDisconnect();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnReceivedMessage(LPCTSTR message)
Date    : 16.12.97
-------------------------------------------------------------------------------
Parameter: message that has been received from the server
Returnval: void
Description: calls the TPerception ParseMessage method to parse the message
				 calls the views display message function
-----------------------------------------------------------------------------*/
void TViewingAgent::OnReceivedMessage(LPCTSTR message, BOOL bFromNetwork)
{
	// parse the message
	m_pPerception->NewMessage(message);

	// display the message in the message view
	m_pView->Message(message, bFromNetwork);

	// start the timeout timer if init message
	m_pView->SetReceivedData(TRUE);
	if (message[1] == 'i')
		m_pView->StartTimeoutTimer();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnSentMessage(LPCTSTR message)
Date    : 17.12.97
-------------------------------------------------------------------------------
Parameter: message that has been sent to the server
Returnval: (void)
Description: displays the message in the message view
-----------------------------------------------------------------------------*/
void TViewingAgent::OnSentMessage(LPCTSTR message)
{
	m_pView->Message(message, TRUE);
}


/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnMoved(int x, int y)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: position the player tried to move to
Returnval: (void)
Description: called after the agent performed a move command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnMoved(int x, int y)
{
	// update the position of this player
	TPlayer*	pPlayer;

	pPlayer = m_pPerception->GetThisPlayer();
	if (GetSide() == RIGHT_TO_LEFT)
	{
		pPlayer->SetAbsX(-x);
		pPlayer->SetAbsY(-y);
	}
	else
	{
		pPlayer->SetAbsX(x);
		pPlayer->SetAbsY(y);
	}
	// tell the player, that all the other info is out of date
	m_pPerception->InvalidateAll();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnDashed(int power)
Date    : 27.08.98
-------------------------------------------------------------------------------
Parameter: power of the dash (-30 - 100)
Returnval: (void)
Description: called after a dash command. updates the stamina of the player
-----------------------------------------------------------------------------*/
void TViewingAgent::OnDashed(int power)
{
	m_pPerception->OnDashed(power);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnKicked(int power, int direction)
Date    : 03.09.98
-------------------------------------------------------------------------------
Parameter: power and relative direction of the kick command
Returnval: (void)
Description: adjusts the ball's speed
-----------------------------------------------------------------------------*/
void TViewingAgent::OnKicked(int power, int direction)
{
	m_pPerception->OnKicked(power, direction);	
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnTurned(int dir)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: moment with which the player turned. this may be more than the player turned
			  actually, especially if the player has got high speed (!!!)
Returnval: (void)
Description: called after the agent performed a turn command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnTurned(int dir)
{
	double	speed = m_pPerception->GetSpeed();

	// turn has got some inertia, so actual turn is less than dir
	double	dTurn = dir;
	dTurn /= 1.0 + speed * INERTIA_MOMENT;
	dir = (int)dTurn;

	// update the direction of this player
	m_pPerception->OnTurned(dir);
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnCaught(int dir)
Date    : 29.07.98
-------------------------------------------------------------------------------
Parameter: direction of the catch command
Returnval: (void)
Description: called, when the agent executed a catch command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnCaught(int dir)
{
	m_pPerception->SetLastCatchTime();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::OnTurnedNeck(int dir)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: angle with which the player turned the neck. 
Returnval: (void)
Description: called after the agent performed a turn neck command
-----------------------------------------------------------------------------*/
void TViewingAgent::OnTurnedNeck(int dir)
{
	m_pPerception->OnTurnedNeck(dir);
}


/*-----------------------------------------------------------------------------
Function: functions passed to TPerception and TBehaviorNetwork
Date    : 26.03.98
-----------------------------------------------------------------------------*/
int TViewingAgent::GetSide()
{
	return m_pPerception->GetSide();
}

void TViewingAgent::SetSide(int side)
{
	m_pPerception->SetSide(side);
}

int TViewingAgent::GetStamina() const
{
	return m_pPerception->GetStamina();
}

void TViewingAgent::ResetNetwork()
{
	m_pBehaviorNetwork->ResetNetwork();
}

TNetworkParams* TViewingAgent::GetNetworkParams()
{
	return m_pBehaviorNetwork->GetNetworkParams();
}

int TViewingAgent::GetNoOfRules() const
{
	return m_pBehaviorNetwork->GetNoOfRules();
}


/*-----------------------------------------------------------------------------
Function: double TViewingAgent::GetLinearFuzzyValue(double value, double left, double right, int up)
Date    : 13.03.98
-------------------------------------------------------------------------------
Parameter: beginning point of increase (decrease); values less than left return 0.0 (1.0)
			  end point of increase (decrease); values greater than right return 1.0 (0.0)
Returnval: fuzzy value 0.0 <= val <= 1.0
Description: calculates a fuzzy value of value as linear increase (up = TRUE default)
				 or decrease (up = FALSE) between left and right border
-----------------------------------------------------------------------------*/
double TViewingAgent::GetLinearFuzzyValue(double value, double left, double right, int up)
{
	ASSERT(left < right);

	if (up)
		if (value < left)
			return 0.0;
		else if (value > right)
			return 1.0;
		else
			return (value - left) / (right - left);
	else
		if (value < left)
			return 1.0;
		else if (value > right)
			return 0.0;
		else
			return (right - value) / (right - left);
}

	
/*-----------------------------------------------------------------------------
Function: void TViewingAgent::CreatePerceptions()
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: adds the perception functions to the behavior networks function factory
-----------------------------------------------------------------------------*/
void TViewingAgent::CreatePerceptions()
{
	m_pBehaviorNetwork->AddPerception("ballAhead", this, (PCallBackPerception) GetBallAhead);
	m_pBehaviorNetwork->AddPerception("ballBehind", this, (PCallBackPerception) GetBallBehind);
	m_pBehaviorNetwork->AddPerception("ballIsCatchable", this, (PCallBackPerception) GetBallIsCatchable);
	m_pBehaviorNetwork->AddPerception("ballIsFar", this, (PCallBackPerception) GetBallIsFar);
	m_pBehaviorNetwork->AddPerception("ballIsNearOtherGoal", this, (PCallBackPerception) GetBallIsNearOtherGoal);
	m_pBehaviorNetwork->AddPerception("ballIsNearOwnGoal", this, (PCallBackPerception) GetBallIsNearOwnGoal);
	m_pBehaviorNetwork->AddPerception("ballIsNearOwnGoal", this, (PCallBackPerception) GetBallIsNearOwnGoal);
	m_pBehaviorNetwork->AddPerception("ballIsVeryNearOwnGoal", this, (PCallBackPerception) GetBallIsVeryNearOwnGoal);
	m_pBehaviorNetwork->AddPerception("ballKickable", this, (PCallBackPerception) GetBallKickable);
	m_pBehaviorNetwork->AddPerception("ballOutside", this, (PCallBackPerception) GetBallOutside);
	m_pBehaviorNetwork->AddPerception("catchSuccess", this, (PCallBackPerception) GetCatchSuccess);
	m_pBehaviorNetwork->AddPerception("freeTeammateIsOffside", this, (PCallBackPerception) GetFreeTeammateIsOffside);
	m_pBehaviorNetwork->AddPerception("haveBall", this, (PCallBackPerception) GetHaveBall);
	m_pBehaviorNetwork->AddPerception("haveEnoughStamina", this, (PCallBackPerception) GetHaveEnoughStamina);
	m_pBehaviorNetwork->AddPerception("haveSeenBall", this, (PCallBackPerception) GetHaveSeenBall);
	m_pBehaviorNetwork->AddPerception("highStamina", this, (PCallBackPerception) GetHighStamina);
	m_pBehaviorNetwork->AddPerception("inOffside", this, (PCallBackPerception) GetInOffside);
	m_pBehaviorNetwork->AddPerception("inOtherHalf", this, (PCallBackPerception) GetInOtherHalf);
	m_pBehaviorNetwork->AddPerception("justCaughtBall", this, (PCallBackPerception) GetJustCaughtBall);
	m_pBehaviorNetwork->AddPerception("knowNotInOffside", this, (PCallBackPerception) GetKnowNotInOffside);
	m_pBehaviorNetwork->AddPerception("lowStamina", this, (PCallBackPerception) GetLowStamina);
	m_pBehaviorNetwork->AddPerception("maySaySomething", this, (PCallBackPerception) GetMaySaySomething);
	m_pBehaviorNetwork->AddPerception("nearBall", this, (PCallBackPerception) GetNearBall);
	m_pBehaviorNetwork->AddPerception("nearGoal", this, (PCallBackPerception) GetNearGoal);
	m_pBehaviorNetwork->AddPerception("nearOpponent", this, (PCallBackPerception) GetNearOpponent);
	m_pBehaviorNetwork->AddPerception("nearOwnGoal", this, (PCallBackPerception) GetNearOwnGoal);
	m_pBehaviorNetwork->AddPerception("noOtherGoal", this, (PCallBackPerception) GetNoOtherGoal);
	m_pBehaviorNetwork->AddPerception("offsideTrap", this, (PCallBackPerception) GetOffsideTrap);
	m_pBehaviorNetwork->AddPerception("opponentInFront", this, (PCallBackPerception) GetOpponentInFront);
	m_pBehaviorNetwork->AddPerception("playmodeOn", this, (PCallBackPerception) GetPlaymodeOn);
	m_pBehaviorNetwork->AddPerception("soccergoal", this, (PCallBackPerception) GetSoccergoal);
	m_pBehaviorNetwork->AddPerception("standFree", this, (PCallBackPerception) GetStandFree);
	m_pBehaviorNetwork->AddPerception("teammateBehind", this, (PCallBackPerception) GetTeammateBehind);
	m_pBehaviorNetwork->AddPerception("teammateHasBall", this, (PCallBackPerception) GetTeammateHasBall);
	m_pBehaviorNetwork->AddPerception("teammateInFront", this, (PCallBackPerception) GetTeammateInFront);
	m_pBehaviorNetwork->AddPerception("teammateIsBetweenBall", this, (PCallBackPerception) GetTeammateIsBetweenBall);
	m_pBehaviorNetwork->AddPerception("teammateIsFree", this, (PCallBackPerception) GetTeammateIsFree);
	m_pBehaviorNetwork->AddPerception("teammateIsNearerBall", this, (PCallBackPerception) GetTeammateIsNearerBall);
}


/*-----------------------------------------------------------------------------
Function: Perception functions
Date    : 14.01.98
-------------------------------------------------------------------------------
Parameter: do not get parameters
Returnval: activation (truth) of this perception
Description: this are the functions that calculate the truth of perceptions
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
returns true, if the ball is nearer to the other goal (only x coordinate)
than this player is
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallAhead()
{
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	x2 = (int) m_pPerception->GetBall()->GetAbsX();

	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if ((x2 < 47) && (x2 > x1))
			return GetLinearFuzzyValue(x2 - x1, 7.0, 20.0);
		else
			return 0.0;
	}
	else if ((x2 > -47) && (x2 < x1))
			return GetLinearFuzzyValue(x1 - x2, 7.0, 20.0);
		else
			return 0.0;
}

/*-----------------------------------------------------------------------------
returns true, if the ball is nearer to the own goal (only x coordinate)
than this player is
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallBehind()
{
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	x2 = (int) m_pPerception->GetBall()->GetAbsX();

	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if (x2 > -47)
			return GetLinearFuzzyValue(x1 - x2, -4.0, 0.0);
		else
			return 0.0;
	}
	else if (x2 < 47)
			return GetLinearFuzzyValue(x2 - x1, -4.0, 0.0);
		else
			return 0.0;
}

/*-----------------------------------------------------------------------------
calculates if the ball is or will be within reach to catch (goalie)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsCatchable()
{
	TVector			ballInFuture;
	TVector			toBall;
	TBall*			pBall = m_pPerception->GetBall();
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();
	double			ballDist;

	// get distance of ball next cycle
	pBall->GetFuturePosition(1, ballInFuture);
	ballDist = pAgent->GetDistanceTo(ballInFuture);

	if (ballDist > 2.0)
		return 0.0;
	else if (ballDist > 1.8)
		return 0.4;
	else
		return 1.0;
}

/*-----------------------------------------------------------------------------
false up to 15 m of ball distance to agent, true at 25 or greater distance
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsFar()
{
	if (!m_pPerception->GetBall()->IsInMind())
		return 0.0;
	else
		return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(m_pPerception->GetBall()), 15.0, 25.0);
}

/*-----------------------------------------------------------------------------
perception proposition for the distance of the ball to the other goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsNearOtherGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetBall()->GetDistanceTo(
											m_pPerception->GetOtherGoal()), 15.0, 70.0, FALSE);
}

/*-----------------------------------------------------------------------------
perception proposition for the distance of the ball to the own goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsNearOwnGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetBall()->GetDistanceTo(
											m_pPerception->GetOwnGoal()), 15.0, 70.0, FALSE);
}

/*-----------------------------------------------------------------------------
perception proposition for the (future) distance of the ball to the own goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallIsVeryNearOwnGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetOwnGoal()
										->GetDistanceTo(*(m_pPerception->GetFutureBallPosition(10)))
										, 14.0, 25.0, FALSE);
}

/*-----------------------------------------------------------------------------
true if the ball is within the kickable_area of this agent
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallKickable()
{
	TVector	ballNow;
	TVector	ballSpeed;
	TVector	intersection;
	TVector	line;
	TVector	orthogonal;
	double	dMinDist = 10000.0;
	double	maxKickDistance = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;	
	double	s = 0.0;

	// calculate the trajectory the ball takes
	ballNow = m_pPerception->GetBall()->GetPosition();
	ballSpeed = m_pPerception->GetBall()->GetAbsSpeed();
	ballSpeed -= m_pPerception->GetThisPlayer()->GetAbsSpeed();

	// calculate closest point to ball trajectory 
	intersection = ballNow;
	if (ballSpeed.GetLength() > 0.1)
	{
		// ball has significant speed
		double	px, py;		// ball coordinates
		double	lx, ly;		// direction of ball speed
		double	rx, ry;		// position of this agent
		double	ratio;		// oy/ox where vector o is orthogonal to line

		px = ballNow.GetX();
		py = ballNow.GetY();
		lx = ballSpeed.GetX();
		ly = ballSpeed.GetY();
		rx = m_pPerception->GetThisPlayer()->GetAbsX();
		ry = m_pPerception->GetThisPlayer()->GetAbsY();
		
		if ((ly > D_ZERO_N) && (ly < D_ZERO_P))
			// the ball is approaching us
			s = (rx - px) / lx;
		else
		{
			ratio = -lx/ly;		// is orthogonal to trajectory of ball
			s = (py - px * ratio - ry + rx * ratio) / (lx * ratio - ly);
		}
		
		#ifdef SHOW_VIEWING_VAR
			TRACE1("BallKickable in : %4.2f\n", s);
		#endif

		if (s > 2.0)
			// just look 2 timesteps into future
			s = 2.0;
		
		if (s > 0.0)
			// take the intersection point to evaluate kickability
			// in all other cases the current ball position is taken
			intersection = ballNow + s * ballSpeed;
	}
	
	// get the distance to the intersection point
	dMinDist = m_pPerception->GetThisPlayer()->GetDistanceTo(intersection);
		
	if (dMinDist > maxKickDistance)
		return 0.0;
	else 
	{
		#ifdef SHOW_VIEWING_VAR
			TRACE2("ballKickable: distance %4.2f  s: %4.2f\n", dMinDist, s);
		#endif
		return GetHaveSeenBall();
	}
}

/*-----------------------------------------------------------------------------
true at kick_in or corner_kick
-----------------------------------------------------------------------------*/
double TViewingAgent::GetBallOutside()
{
	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if ((m_pPerception->GetPlayMode() == KICK_IN_L)
			|| (m_pPerception->GetPlayMode() == CORNER_KICK_L))
			// we have kick in
			return 1.0;
	}
	else
	{
		if ((m_pPerception->GetPlayMode() == KICK_IN_R)
			|| (m_pPerception->GetPlayMode() == CORNER_KICK_R))
			// we have kick in
			return 1.0;
	}
	return 0.0;
}

/*-----------------------------------------------------------------------------
true, immediately after a successfull catch
-----------------------------------------------------------------------------*/
double TViewingAgent::GetCatchSuccess()
{
	if (m_pPerception->GetGoalieCaught())
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true, if the free teammate is in offside
-----------------------------------------------------------------------------*/
double TViewingAgent::GetFreeTeammateIsOffside()
{
	TPlayer*	pTeammate = m_pPerception->GetFreeTeammate();
	
	if (!GetOrdinaryGame() && (pTeammate != NULL))
	{
		// check if the player is offside and if so, do not pass to him
		double	dTeammateX = pTeammate->GetAbsX();
		double	dBallX = m_pPerception->GetBall()->GetAbsX();
		TPlayer*	pOpp = m_pPerception->Get2BackmostOpponent();

		if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
		{
			// check if teammate is in other side and before ball
			if ((dTeammateX > 0) && (dTeammateX > dBallX)
				&& ((pOpp == NULL) || (pOpp->GetAbsX() < dTeammateX)))
			{
				// the teammate is offside
				return 1.0;
			}
		}
		else
		{
			// check if teammate is in other side and before ball
			if ((dTeammateX < 0) && (dTeammateX < dBallX)
				&& ((pOpp == NULL) || (pOpp->GetAbsX() > dTeammateX)))
			{
				// the teammate is offside
				return 1.0;
			}
		}
	}
	return 0.0;
}

/*-----------------------------------------------------------------------------
true, if this agent is nearest to ball and near enough (but maybe farer than ballKickable)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHaveBall()
{
	TPlayer*		pPlayerAtBall = m_pPerception->GetAnyPlayerAtBall();
	double		ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());

	// is agent nearest to Ball
	if (pPlayerAtBall != NULL)
	{
		// look, if the other agent is nearer to ball
		if (pPlayerAtBall->GetDistanceTo(m_pPerception->GetBall()) < ballDist)
			// the other player is nearer to ball
			return 0.0;
	}

	return GetLinearFuzzyValue(ballDist, 1.0, 9.0, FALSE) * GetHaveSeenBall();
}

/*-----------------------------------------------------------------------------
true, if this agent has enough stamina to dash 100% to the ball
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHaveEnoughStamina()
{
	double	ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	int		stamina = m_pPerception->GetStamina();

	return GetLinearFuzzyValue(stamina - ballDist * 85, 0, LOW_STAMINA);
}

/*-----------------------------------------------------------------------------
true, if the ball has been seen within the last 3 seconds
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHaveSeenBall()
{
	// if the ball is not in mind, it has not been seen
	if (!m_pPerception->GetBall()->IsInMind())
		return 0.0;
	
	int deltaTime = m_pPerception->GetSeeTime() - m_pPerception->GetObjectTime(m_pPerception->GetBall());
	return GetLinearFuzzyValue((double)deltaTime, 0.0, (double) MAX_IMAGINE_TIME, FALSE);	
}

/*-----------------------------------------------------------------------------
true if the player has got enough power to dash 
-----------------------------------------------------------------------------*/
double TViewingAgent::GetHighStamina()
{
	//!!!(must be tuned)
	return GetLinearFuzzyValue(m_pPerception->GetStamina(), 1500.0, 2000.0);
}

/*-----------------------------------------------------------------------------
true, if the agent knows he is in offside
-----------------------------------------------------------------------------*/
double TViewingAgent::GetInOffside()
{
	 return m_pPerception->GetInOffside();
}

/*-----------------------------------------------------------------------------
true if the agent is in the opponents half
-----------------------------------------------------------------------------*/
double TViewingAgent::GetInOtherHalf()
{
	if (m_pPerception->GetThisPlayer()->IsInOwnHalfOfField())
		return 0.0;
	else
		return 1.0;
}

/*-----------------------------------------------------------------------------
true, if the goalie caught the ball within the last 5 cycles. A goalie can only
catch the ball once each 5 cycles
-----------------------------------------------------------------------------*/
double TViewingAgent::GetJustCaughtBall()
{
	int	lastCatch = m_pPerception->GetSenseTime() - m_pPerception->GetLastCatchTime();
	if (lastCatch < 6)
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if the player knows he is not in offside
-----------------------------------------------------------------------------*/
double TViewingAgent::GetKnowNotInOffside()
{
	if (m_pPerception->GetKnowNotInOffside())
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if the player has got low power to dash 
-----------------------------------------------------------------------------*/
double TViewingAgent::GetLowStamina()
{
	int	iStart = LOW_STAMINA;
	int	iTime = m_pPerception->GetSenseTime();

	if ((iTime > 5500) && (iTime < 6000))
		// end of second half
		iTime -= 3000;

	if ((iTime > 2500) && (iTime < 3000))
		// end of first or second half
		iStart = 500;

	return GetLinearFuzzyValue(m_pPerception->GetStamina(), iStart, iStart + 200, FALSE);
}

/*-----------------------------------------------------------------------------
true if the player is on turn to say something
-----------------------------------------------------------------------------*/
double TViewingAgent::GetMaySaySomething()
{
	int	time = m_pPerception->GetSenseTime();

	// a player may say something each 22 cycle
	// the players say in order 1 2 3 4 ...
//	if ((m_pPerception->GetTimeProceeding()) && !(time % 2) && (((time / 2) % 11) + 1 == m_iNumber))
	// the players say in order 1 5 9 2 6 10 3 7 11 4 8
	if ((m_pPerception->GetTimeProceeding()) && !(time % 2) 
			&& (((time * 2) % 11) + 1 == m_iNumber))
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if this player is close to the ball
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearBall()
{
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(m_pPerception->GetBall())
										, 5.0, 25.0, FALSE)
			 * GetHaveSeenBall();
}

/*-----------------------------------------------------------------------------
true if this player is close to the ball
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(m_pPerception->GetOtherGoal())
										, 10.0, 20.0, FALSE);
}

/*-----------------------------------------------------------------------------
true if this player is close to an opponent
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearOpponent()
{
	TPlayer*	player = m_pPerception->GetOpponentAtMe();
	if (player == NULL)
		return 0.0;
	
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(player)
										, 3.0, 7.0, FALSE);
}

/*-----------------------------------------------------------------------------
true if this player is close to the own goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNearOwnGoal()
{
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(m_pPerception->GetOwnGoal())
										, 10.0, 30.0, FALSE);
}

/*-----------------------------------------------------------------------------
only false in the case that the opponent scored a goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetNoOtherGoal()
{
	if (((m_pPerception->GetSide() == LEFT_TO_RIGHT) 
			&& (m_pPerception->GetPlayMode() == GOAL_R)) 
		|| ((m_pPerception->GetSide() == RIGHT_TO_LEFT)
			&& (m_pPerception->GetPlayMode() == GOAL_L)))
		return 0.0;
	else
		return 1.0;
}

/*-----------------------------------------------------------------------------
just for goal offsideTrap
-----------------------------------------------------------------------------*/
double TViewingAgent::GetOffsideTrap()
{
	return 0.0;
}

/*-----------------------------------------------------------------------------
true if an opponent is between this agent and the opponents goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetOpponentInFront()
{
	// is there a player in the direction to the other goal
	// !!! this has to be another fuction than getplayerinway
	TPlayer* pPlayer = m_pPerception->GetPlayerInWay(m_pPerception->GetObjectDirection(
								m_pPerception->GetOtherGoal()));
	if (pPlayer == NULL)
		return 0.0;

	// there is a player
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(pPlayer), 6.0, 15.0, FALSE);
}

/*-----------------------------------------------------------------------------
true if playmode is play_on
-----------------------------------------------------------------------------*/
double TViewingAgent::GetPlaymodeOn()
{
	if (m_pPerception->GetPlayMode() == PLAY_ON) 
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if a goal has been scored this cycle
-----------------------------------------------------------------------------*/
double TViewingAgent::GetSoccergoal()
{
	if (((m_pPerception->GetSide() == LEFT_TO_RIGHT) 
			&& (m_pPerception->GetPlayMode() == GOAL_L)) 
		|| ((m_pPerception->GetSide() == RIGHT_TO_LEFT)
			&& (m_pPerception->GetPlayMode() == GOAL_R)))
		return 1.0;
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true, if this agent is unmarked i.e. the ball can be passed to him
-----------------------------------------------------------------------------*/
double TViewingAgent::GetStandFree()
{
	// opposite to GetNearOpponent
	TPlayer*	player = m_pPerception->GetOpponentAtMe();
	if (player == NULL)
		return 1.0;
	
	return GetLinearFuzzyValue(m_pPerception->GetObjectDistance(player), 3.0, 8.0);
}

/*-----------------------------------------------------------------------------
true if a teammate is behind me in x direction (used for offside trap)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateBehind()
{
	TPlayer*	pPlayer = m_pPerception->Get2BackmostTeammate();

	if (pPlayer != NULL)
	{
		double	dTeammateX = pPlayer->GetAbsX();
		double	dAgentX = m_pPerception->GetThisPlayer()->GetAbsX();

		if (m_pPerception->GetSide() == RIGHT_TO_LEFT)
			dTeammateX -= dAgentX;
		else
			dTeammateX = dAgentX - dTeammateX;
	
		return GetLinearFuzzyValue(dTeammateX, 4.0, 9.0);
	}
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if the teammate is the closest player to the ball and is close enough
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateHasBall()
{
	TPlayer*	pPlayer = m_pPerception->GetAnyPlayerAtBall();

	if ((pPlayer != NULL) && (pPlayer->IsOwnTeam()))
		return GetLinearFuzzyValue(pPlayer->GetDistanceTo(m_pPerception->GetBall()), 1.0, 9.0, FALSE);
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
true if a teammate is in front of me in x direction (used for offside prevention)
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateInFront()
{
	TPlayer*	pPlayer = m_pPerception->GetFrontmostTeammate();

	if (pPlayer != NULL)
	{
		double	dTeammateX = pPlayer->GetAbsX();
		double	dAgentX = m_pPerception->GetThisPlayer()->GetAbsX();

		if (m_pPerception->GetSide() == RIGHT_TO_LEFT)
			dTeammateX = dAgentX - dTeammateX;
		else
			dTeammateX -= dAgentX;
	
		return GetLinearFuzzyValue(dTeammateX, 2.0, 5.0);
	}
	else
		return 0.0;
}

/*-----------------------------------------------------------------------------
prevents the goalie from running to the ball if a teammate is between him and the ball.
The goalie has to run to the ball even if he is not the closest teammate to the ball.
This is the case, if an opponent followed by a teammate is running to the goal
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateIsBetweenBall()
{
	// checks, if a teammate is nearer to the ball and between me and the ball
	double	dNearer = GetTeammateIsNearerBall();

	if (dNearer > 0.5)
	{
		// teammate is nearer, so check if he is nearer to the own goal as
		// opponent at ball is
		TPlayer*				ownPlayer = m_pPerception->GetTeammateAtBall();
		TPlayer*				opponent = m_pPerception->GetOpponentAtBall();
		TVisibleObject*	ownGoal = m_pPerception->GetOwnGoal();
		double				deltaDist;

		if ((ownPlayer == NULL) || (opponent == NULL))
			return dNearer;

		deltaDist = opponent->GetDistanceTo(ownGoal) - ownPlayer->GetDistanceTo(ownGoal);
		return GetLinearFuzzyValue(deltaDist, 1.0, 3.0) * dNearer;
	}
	else
		return dNearer;
}

/*-----------------------------------------------------------------------------
true if one of the teammates is unmarked
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateIsFree()
{
	TPlayer*		pFreeTeammate = m_pPerception->GetFreeTeammate();

	if (pFreeTeammate == NULL)
		return 0.0;
	else
		return pFreeTeammate->GetFreePassDegree();
}


/*-----------------------------------------------------------------------------
true if a teammate is closer to the ball than this player
-----------------------------------------------------------------------------*/
double TViewingAgent::GetTeammateIsNearerBall()
{
	// that player is nearer ball, that will be faster at the balls future position
	TPlayer*		ownPlayer = m_pPerception->GetTeammateAtBall();
	double		deltaDist;
	TVector		ballInFuture;

	ballInFuture = *m_pPerception->GetFutureBallPosition(7);
	if ((ownPlayer == NULL) || (!ownPlayer->IsInMind()))
		// if there is no one playing with me
		return 0.0;
	
	deltaDist = m_pPerception->GetThisPlayer()->GetDistanceTo(ballInFuture)
					- ownPlayer->GetDistanceTo(ballInFuture);

	return GetLinearFuzzyValue(deltaDist, 0.0, 1.0)
			 * GetHaveSeenBall();
}

/*-----------------------------------------------------------------------------
Function: void TViewingAgent::CreateResources()
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: adds the perception functions to the behavior networks function factory
-----------------------------------------------------------------------------*/
void TViewingAgent::CreateResources()
{
	m_pBehaviorNetwork->AddResource("leg", this, (PCallBackPerception) GetLegsAvailable, 1);
	m_pBehaviorNetwork->AddResource("mouth", this, (PCallBackPerception) GetMouthsAvailable, 1);
	m_pBehaviorNetwork->AddResource("neck", this, (PCallBackPerception) GetNecksAvailable, 1);
	m_pBehaviorNetwork->AddResource("stamina", this, (PCallBackPerception) GetStaminaAvailable, MAX_STAMINA);
}	


double TViewingAgent::GetLegsAvailable()
{
	return 1.0;
}

double TViewingAgent::GetMouthsAvailable()
{
	return 1.0;
}

double TViewingAgent::GetNecksAvailable()
{
	return 1.0;
}

double TViewingAgent::GetStaminaAvailable()
{
	return m_pPerception->GetStamina();
}
